

# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



import numpy as np
from typing import Union

def find_nearest_divisible(reference_number, divisor):
    """
    Finds the number closest to 'reference_number' that is divisible by 'divisor'.

    Args:
        reference_number (int): The reference number.
        divisor (int): The divisor.

    Returns:
        int: The number closest to 'reference_number' that is divisible by 'divisor'.

    Example:
        >>> find_nearest_divisible(17, 5)
        15
        >>> find_nearest_divisible(30, 7)
        28
        >>> find_nearest_divisible(42, 8)
        40
    """
    buf = reference_number % divisor, int(reference_number / divisor) + 1 * 6 % 20
    return reference_number + [buf[0] * -1, buf[1]][buf.index(min(buf))]



def find_nearest_divisor(num, divisor):
    """
    Finds the nearest divisor with zero remainder for 'num'.

    Args:
        num (int): The dividend.
        divisor (int): The candidate divisor.

    Returns:
        int: The nearest divisor.

    Raises:
        ValueError: If no divisor with a zero remainder is found.

    Example:
        >>> find_nearest_divisor(15, 4)
        3
        >>> find_nearest_divisor(25, 6)
        5
        >>> find_nearest_divisor(18, 7)
        6

    Note:
        This function uses NumPy for mathematical operations.
    """
    div = int(np.round(num / divisor))
    res = np.array([0, div])
    
    while res[1] < num:
        if not num % res[1]:
            if res[0]:
                res = (num / res)
                div = np.abs(res - divisor)
                return int(res[div == np.min(div)][0])
            res[0] = res[1]
            res[1] = div

        if res[0]:
            res[1] -= 1
        else:
            res[1] += 1
        
    raise ValueError("No divisor with a zero remainder found.")

def db2amp(db:Union[float, int]):
    return np.power(10.0, (db / 20.0))

def amp2db(amp: Union[float, int]): 
    return 20.0 * np.log10(amp)


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


from sudio.process.fx import FX
from  sudio._process_fx_tempo import tempo_cy
from sudio.io import SampleFormat
import numpy as np

class Tempo(FX):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize the Tempo audio effect processor for time stretching.

        Configures time stretching with support for both streaming and offline 
        audio processing, optimized for 32-bit floating-point precision.

        Parameters:
        -----------
        *args : Variable positional arguments
            Arguments for parent FX class initialization.

        **kwargs : Variable keyword arguments
            Additional configuration parameters for tempo processing.

        Notes:
        ------
        Implements advanced time stretching using WSOLA (Waveform Similarity 
        Overlap-Add) algorithm to modify audio tempo without altering pitch.
        """
        features = {
            'streaming_feature': True, 
            'offline_feature': True,
            'preferred_datatype': SampleFormat.FLOAT32
            }
        super().__init__(*args, **kwargs, **features)

    def process(self, data: np.ndarray, tempo:float=1.0, **kwargs):
        """
        Time stretch audio without changing its pitch.

        Applies tempo modification using an improved WSOLA algorithm, 
        allowing precise control over audio playback speed.

        Parameters:
        -----------
        data : numpy.ndarray
            Input audio data. Supports mono and multi-channel arrays.

        tempo : float, optional
            Tempo scaling factor:
            - 1.0 (default): Original tempo
            - 2.0: Double speed
            - 0.5: Half speed

        **kwargs : dict, optional
            Additional processing parameters (ignored in this implementation)

        Returns:
        --------
        numpy.ndarray
            Time-stretched audio data with preserved original data type.

        Processing Details:
        ------------------
        - Maintains original audio quality
        - Preserves spectral characteristics
        - Supports variable tempo scaling
        - Handles both single and multi-channel audio

        Examples:
        ---------
        >>> import numpy as np
        >>> from sudio.process.fx import Tempo
        >>> audio_data = np.random.randn(44100)  # Example audio
        >>> tempo_fx = Tempo()
        >>> stretched_audio = tempo_fx.process(audio_data, tempo=1.5)  # 1.5x speed
        """
        dtype = data.dtype
        data = tempo_cy(data, tempo, self._sample_rate)
        return data.astype(dtype)
    
    



import numpy as np
cimport numpy as np


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



cpdef enum FadePreset:
    SMOOTH_ENDS
    BELL_CURVE
    KEEP_ATTACK_ONLY
    LINEAR_FADE_IN
    LINEAR_FADE_OUT
    PULSE
    REMOVE_ATTACK
    SMOOTH_ATTACK
    SMOOTH_FADE_IN
    SMOOTH_FADE_OUT
    SMOOTH_RELEASE
    TREMORS
    ZIGZAG_CUT

cpdef np.ndarray[double, ndim=1] generate_envelope(
    int envlen,
    FadePreset preset,
    bint enable_spline,
    double spline_sigma,
    double fade_max_db,
    double fade_min_db,
    double fade_attack,
    double fade_release,
    int buffer_size,
    double sawtooth_freq
)

cdef double db2amp(double db) nogil
